#
#  This file is part of ProfileView.
#
#  ProfileView is free software: you can redistribute it and/or modify
#  it under the terms of the CeCILL 2.1 Licence
#
#  ProfileView is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  You should have received a copy of the Licence CeCILL 2.1 along
#  with ProfileView. If not, see <https://cecill.info/>.
#

import math
import operator

from pv_utils import *


class hmm_model():

    def __init__(self):
        self.name    = ""
        self.length  = 0
        self.nseq    = 0
        self.states  = []
        self.cons    = ""
        self.aa_list = [ c for c in "ACDEFGHIKLMNPQRSTVWY" ]
        self.aa2idx  = { c:i for i,c in enumerate(self.aa_list) }
        self.bg_freq = [ -math.log(0.05) for c in self.aa_list ]

    def __str__(self):
        consensus = []
        for k,s in enumerate(self.states):
            aai, minval = min(enumerate(s), key=operator.itemgetter(1))
            f = self.frequency(k, self.aa_list[aai])
            c = self.aa_list[aai].upper() if f >= .6 else (self.aa_list[aai].lower() if f >= .4 else 'x')
            consensus.append(c)
        return "NAME={}\nLENG={} STATES={}\nNSEQ={}\nCONS={}".format(self.name, self.length,len(self.states),self.nseq, ''.join(consensus))

    def frequency(self, k, aa):
        aaidx = self.aa2idx[aa]
        return math.exp(-self.states[k][aaidx])

#    def consensus(self):
#        if not self.cons:
#            consensus = []
#            for k,s in enumerate(self.states):
#                aai, minval = min(enumerate(s), key=operator.itemgetter(1))

    def getStateFreqDict(self, k):
        outDict = {}
        for aa in aa2idx:
            aaidx = self.aa2idx[aa]
            outDict[aa] = math.exp(-self.states[k][aaidx])
        return outDict

    def ln_odds_ratio(self,k,aa):
        aai   = self.aa2idx[aa]
        m_k_a = self.states[k][aai]
        m_0_a = self.bg_freq[aai]
        return (m_0_a - m_k_a)

    def score(self, k, aa):
        return self.ln_odds_ratio(k,aa)/math.log(2.0) # convert to base2 logarithm

    def is_valid(self):
        return len(self.name) > 0 and self.length > 0 and self.length == len(self.states) and self.nseq > 0


class hmmer_hit():

    def __init__(self):
        self.hmm_name = ""
        self.seq_name = ""
        self.typ      = '?'
        self.score    = 0.0 # HMMER's bit score
        self.mcscore  = 0.0 # sum of scores in well-aligned positions (PP[i] \in {8,9,*}) with a match between the sequence and the hmm consensus
        self.wcscore  = 0.0 # sum of scores in well-aligned positions (PP[i] \in {8,9,*}) with score[i] >= threshold
        self.matches  = 0 # number of matches between the sequence and the hmm consensus
        self.aligned_cols = 0 # number of aligned match states
        self.well_aligned_cols = 0 # number of well-aligned match states
        self.ali_len  = 0 # length of alignment
        self.hmm_from = 0
        self.hmm_to   = 0
        self.ali_from = 0
        self.ali_to   = 0
        self.env_from = 0
        self.env_to   = 0
        self.hmm_cons = ""
        self.seq_cons = ""
        self.pp_cons  = ""
        self.wc_str   = ""
        self.cm_str   = ""
        self.match_str = ""

    def canExtendWith(self,b):
        if self.hmm_name != b.hmm_name or self.seq_name != b.seq_name or self.typ == '?' or b.typ == '?':
            return False
        Lm = b.hmm_to-self.hmm_from+1
        Ls = b.ali_to-self.ali_from+1
        return self.hmm_from <= b.hmm_from and self.ali_from <= b.ali_from and math.fabs(Ls-Lm)/float(min(Ls,Lm)) <= .2

    # it is assumed that canExtendWith(b) holds True
    def extendWith(self,b):
        self.score   += b.score
        self.mcscore += b.mcscore
        self.wcscore += b.wcscore
        self.matches += b.matches
        self.aligned_cols += b.aligned_cols
        self.well_aligned_cols += b.well_aligned_cols
        self.ali_len += b.ali_len
        self.hmm_from = min(self.hmm_from,b.hmm_from)
        self.hmm_to   = max(self.hmm_to,b.hmm_to)
        self.ali_from = min(self.ali_from,b.ali_from)
        self.ali_to   = max(self.ali_to,b.ali_to)
        self.env_from = min(self.env_from,b.env_from)
        self.env_to   = max(self.env_to,b.env_to)
        self.wc_str  += b.wc_str
        self.cm_str  += b.cm_str
        self.match_str += b.match_str
        return self

#    def __str__(self):
#        return "HMM {}\nSEQ {}\nSC  {}\nHMM-SEQ  ALIGNMENT  {}-{}  {}-{}\n{}\n{}\n{}".format(
#                self.hmm_name, self.seq_name, self.score,
#                self.hmm_from, self.hmm_to, self.ali_from, self.ali_to,
#                self.hmm_cons, self.seq_cons, self.pp )


